from contextlib import contextmanager
from infi.projector.plugins import CommandPlugin
from infi.projector.helper import assertions
from infi.projector.helper.utils import open_buildout_configfile, commit_changes_to_buildout
from infi.projector.helper.utils.package_sets import InstallRequiresPackageSet, EggsPackageSet
from textwrap import dedent
from logging import getLogger

logger = getLogger(__name__)

USAGE = """
Usage:
    projector requirements list [--development]
    projector requirements add <requirement> [--development] [--commit-changes]
    projector requirements remove <requirement> [--development] [--commit-changes]
    projector requirements freeze [--with-install-requires] [--newest] [--commit-changes] [--push-changes]
    projector requirements unfreeze [--with-install-requires] [--commit-changes] [--push-changes]


Options:
    requirements list               Show all requirements
    requirements add                add a package to the list of project requirements
    requirements remove             remove a package from project requirement list
    requirements freeze             Creates a versions.cfg file, telling buildout to use specific versions
    requirements unfreeze           Deletes the versions.cfg file, if it exists
    <requirement>                   requirement to add/remove
    --development                   Requirement for the development environment only
    --with-install-requires         Set >= requireements in the install_requires section
    --push-changes                  Push freeze commits
"""

class RequirementsPlugin(CommandPlugin):
    def get_docopt_string(self):
        return USAGE

    def get_command_name(self):
        return 'requirements'

    @assertions.requires_repository
    def parse_commandline_arguments(self, arguments):
        methods = [self.list, self.add, self.remove, self.freeze, self.unfreeze]
        [method] = [method for method in methods
                    if arguments.get(method.__name__)]
        self.arguments = arguments
        method()

    def get_package_set(self):
        return EggsPackageSet() if self.arguments.get("--development", False) else InstallRequiresPackageSet()

    def list(self):
        from pprint import pprint
        pprint(list(self.get_package_set().get()))

    def remove(self):
        package_set = self.get_package_set()
        requirements = package_set.get()
        requirement = self.arguments.get('<requirement>')
        if requirement in requirements:
            requirements.remove(requirement)
            package_set.set(requirements)
        if self.arguments.get("--commit-changes", False):
            message = "remove {} from requirements {}"
            commit_message = message.format(requirement, "(dev)" if self.arguments.get("--development") else '')
            commit_changes_to_buildout(commit_message)

    def add(self):
        package_set = self.get_package_set()
        requirements = package_set.get()
        requirement = self.arguments.get('<requirement>')
        if requirement not in requirements:
            requirements.add(requirement)
            package_set.set(requirements)
        if self.arguments.get("--commit-changes", False):
            message = "adding {} to requirements {}"
            commit_message = message.format(requirement, "(dev)" if self.arguments.get("--development") else '')
            commit_changes_to_buildout(commit_message)

    def freeze(self):
        from infi.projector.helper.utils import freeze_versions, buildout_parameters_context, open_tempfile
        from infi.projector.plugins.builtins.devenv import DevEnvPlugin
        from gitpy import LocalRepository
        from os import curdir
        plugin = DevEnvPlugin()
        plugin.arguments = {'--newest': self.arguments.get("--newest", False)}
        with open_tempfile() as tempfile:
            with buildout_parameters_context(["buildout:update-versions-file={0}".format(tempfile)]):
                plugin.build()
            with open(tempfile) as fd:
                content = fd.read()
            with open(tempfile, 'w') as fd:
                fd.write("[versions]\n" + content)
            freeze_versions(tempfile, self.arguments.get("--with-install-requires", False))
        if self.arguments.get("--commit-changes", False):
            repository = LocalRepository(curdir)
            repository.add("buildout.cfg")
            repository.commit("Freezing dependencies")
        push_changes = self.arguments.get("--push-changes", False)
        if push_changes:
            repository._executeGitCommandAssertSuccess("git push")

    def unfreeze(self):
        from infi.projector.helper.utils import unfreeze_versions
        from gitpy import LocalRepository
        from os import curdir
        unfreeze_versions(self.arguments.get("--with-install-requires", False))
        if self.arguments.get("--commit-changes", False):
            repository = LocalRepository(curdir)
            repository.add("buildout.cfg")
            repository.commit("Unfreezing dependencies")
        push_changes = self.arguments.get("--push-changes", False)
        if push_changes:
            repository._executeGitCommandAssertSuccess("git push")



from os import path, name, curdir
from gitpy import LocalRepository
from infi.pyutils.decorators import wraps

from logging import getLogger
logger = getLogger(__name__)

def is_windows():
    return name == 'nt'

def is_executable_exists(filepath):
    return path.exists("{}.exe".format(filepath) if is_windows() else filepath)

def assert_buildout_executable_exists():
    if not is_executable_exists(path.join("bin", "buildout")):
        logger.error("buildout executable does not exist, run `projector devenv build`")
        raise SystemExit(1)

def assert_setup_py_exists():
    if not path.exists("setup.py"):
        logger.error("setup.py does not exist, run `projector devenv build`")
        raise SystemExit(1)

def assert_buildout_configfile_exists():
    if not path.exists("buildout.cfg"):
        logger.error("buildout.cfg does not exist, the current directory is not a home of a project")
        raise SystemExit(1)

def assert_git_repository():
    if not path.exists(".git"):
        logger.error("the current directory is not a home of a git repository")
        raise SystemExit(1)

def assert_no_uncommitted_changes():
    repository = LocalRepository(curdir)
    changes = repository.getChangedFiles() + repository.getStagedFiles()
    if changes:
        message = "There are changes pending commit, cannot continue. please commit or stash them:\n"
        logger.error(message+repr(changes))
        raise SystemExit(1)

def is_isolated_python_exists():
    return path.exists(path.join("parts", "python", "bin",
                                 "python{}".format('.exe' if is_windows() else '')))

def assert_isolated_python_exists():
    if not is_isolated_python_exists():
        logger.error("Isolated python is required")
        raise SystemExit(1)

def assert_on_branch(branch_name):
    repository = LocalRepository(curdir)
    current_branch = repository.getCurrentBranch()
    if current_branch is None or current_branch.name != branch_name:
        logger.error("not currently on branch {}".format(branch_name))
        raise SystemExit(1)

def is_executable_using_isolated_python(executable_name):
    filepath = path.join("bin", "{}-script.py".format(executable_name) if is_windows() else executable_name)
    with open(filepath) as fd:
        content = fd.read()
    logger.debug("{}:\n{}".format(filepath, content))
    python_relpath = path.normpath("parts/python/bin/python")
    first_line = content.splitlines()[0]
    return first_line.startswith("#!") and python_relpath in first_line

def is_buildout_executable_using_isolated_python():
    return is_executable_using_isolated_python("buildout")

def requires_repository(func):
    @wraps(func)
    def decorator(*args, **kwargs):
        assert_buildout_configfile_exists()
        assert_git_repository()
        return func(*args, **kwargs)
    return decorator

def requires_built_repository(func):
    @wraps(func)
    def decorator(*args, **kwargs):
        assert_buildout_configfile_exists()
        assert_git_repository()
        assert_setup_py_exists()
        return func(*args, **kwargs)
    return decorator

def is_version_tag_exists(version_tag):
    repository = LocalRepository(curdir)
    version_tag = version_tag if version_tag.startswith('v') else 'v' + version_tag
    return version_tag in [tag.name for tag in repository.getTags()]

def assert_version_tag_for_release(version_tag):
    if is_version_tag_exists(version_tag):
        msg = "Version tag {} already released."
        logger.error(msg.format(version_tag))
        raise SystemExit(1)

def assert_version_tag_for_upload(version_tag):
    if all([not is_version_tag_exists(version_tag), version_tag not in ['current', 'latest']]):
        msg = "Version tag {} doesn't exist, cannot upload it."
        logger.error(msg.format(version_tag))
        raise SystemExit(1)

def assert_develop_and_master_not_behind_origin():
    repository = LocalRepository(curdir)
    branches = [repository.getBranchByName(branch_name) for branch_name in ['master', 'develop']]
    branches_with_remote = [branch for branch in branches if branch.getRemoteBranch() is not None]
    for branch in branches_with_remote:
        remote_branch = branch.getRemoteBranch()
        if branch.getMergeBase(remote_branch) != remote_branch:
            logger.error("local branch {} is not on top of origin, please rebase".format(branch))
            raise SystemExit(1)

def assert_develop_branch_on_top_of_master():
    repository = LocalRepository(curdir)
    develop = repository.getBranchByName("develop")
    master = repository.getBranchByName("master")
    if develop.getMergeBase(master) != master:
        logger.error("{} is not on top of {}, please rebase".format(develop, master))
        raise SystemExit(1)



__version__ = "0.6.1"
__git_commiter_name__ = "Guy Rozendorn"
__git_commiter_email__ = "guy@rzn.co.il"
__git_branch__ = '(detached from v0.6.1)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = 'bbe51ab5b2d65cb8ef76399a8f2bdb0531027a0a'
__git_head_subject__ = 'HOSTDEV-1095 new bootstrap'
__git_head_message__ = '================================================================================\n\nHost Infrastructure / HOSTDEV-1095\nbootstrap fails if older zc.buildout is installed\n\nType:              Bug                               Status:        Open                       Assignee: Guy Rozendorn\nPriority:          Minor                             Resolution:    Unresolved                 Reporter: Guy Rozendorn\nAffects Version/s:                                   Fix Version/s: 0.30\nComponents: packaging                                                                          Created: 2013-07-07 16:59\nLabels:                                                                                        Updated: 2013-07-07 16:59\n\nIssue Links:\n\nSub-Tasks:\n\nDescription:\n\nComments:'
__git_dirty_diff__ = ''


from contextlib import contextmanager
from infi.projector.plugins import CommandPlugin
from infi.projector.helper import assertions
from infi.projector.helper.utils import open_buildout_configfile, commit_changes_to_buildout
from infi.projector.helper.utils.package_sets import PackageDataSet
from textwrap import dedent
from logging import getLogger

logger = getLogger(__name__)

USAGE = """
Usage:
    projector isolated-python python-version get
    projector isolated-python python-version set <version> [--commit-changes]

Options:
    <version>               Python version to set
"""

class IsolatedPythonPlugin(CommandPlugin):
    def get_docopt_string(self):
        return USAGE

    def get_command_name(self):
        return 'isolated-python'

    @assertions.requires_repository
    def parse_commandline_arguments(self, arguments):
        methods = [self.python_version]
        [method] = [method for method in methods
                    if arguments.get(method.__name__.replace('_', '-'))]
        self.arguments = arguments
        method()

    def get_package_set(self):
        return PackageDataSet()

    def python_version(self):
        with open_buildout_configfile(write_on_exit=self.arguments.get("set")) as buildout:
            sections = [section for section in buildout.sections()
                        if buildout.has_option(section, "recipe") and \
                        buildout.get(section, "recipe") == "infi.recipe.python"]
            if not sections: # pragma: no cover
                logger.error("isolated python section not found in buildout.cfg")
                raise SystemExit(1)
            if self.arguments.get("get"):
                print buildout.get(sections[0], "version")
            elif self.arguments.get("set"):
                version = self.arguments.get("<version>")
                if not version.startswith("v"):
                    version = "v" + version
                buildout.set(sections[0], "version", version)
        if self.arguments.get("--commit-changes", False):
            commit_message = "changed isolated python version to {}".format(version)
            commit_changes_to_buildout(commit_message)



from sys import argv
#pylint: disable=R0801

def projector(argv=argv[1:]):
    from os import environ
    from logging import basicConfig, INFO, DEBUG, getLogger
    from sys import stderr
    from infi.projector.commandline_parser import parse_commandline_arguments
    basicConfig(level=DEBUG if environ.get("DEBUG") else INFO, stream=stderr)
    getLogger(__name__).debug(' '.join(['projector'] + argv))
    parse_commandline_arguments(argv)




SETUP_INFO = dict(
    name = 'infi.projector',
    version = '0.6.1',
    author = 'Guy Rozendorn',
    author_email = 'guy@rzn.co.il',

    url = 'https://github.com/Infinidat/infi.projector',
    license = 'PSF',
    description = """Python, git-flow based project management tool""",
    long_description = """For the complete document, see the README.md file over at GitHub""",

    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Python Software Foundation License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],

    install_requires = ['infi.execute', 'smmap', 'infi.traceback', 'gitdb', 'infi.exceptools', 'gitflow', 'zc.buildout', 'infi.recipe.template.version', 'setuptools', 'infi.pysync', 'infi.pyutils', 'async', 'git-py', 'docopt'],
    namespace_packages = ['infi'],

    package_dir = {'': 'src'},
    package_data = {'': ['buildout.cfg', '.gitignore', 'README.md', 'bootstrap.py', 'setup.in']},
    include_package_data = True,
    zip_safe = False,

    entry_points = dict(
        console_scripts = ['projector = infi.projector.scripts:projector'],
        gui_scripts = [],
        projector_command_plugins = ['repository = infi.projector.plugins.builtins.repository:RepositoryPlugin', 'envenv = infi.projector.plugins.builtins.devenv:DevEnvPlugin', 'version = infi.projector.plugins.builtins.version:VersionPlugin', 'requirements = infi.projector.plugins.builtins.requirements:RequirementsPlugin', 'console_scripts = infi.projector.plugins.builtins.console_scripts:ConsoleScriptsPlugin', 'gui_scripts = infi.projector.plugins.builtins.gui_scripts:GuiScriptsPlugin', 'package_scripts = infi.projector.plugins.builtins.package_scripts:PackageScriptsPlugin', 'package_data = infi.projector.plugins.builtins.package_data:PackageDataPlugin', 'isolated_pyton = infi.projector.plugins.builtins.isolated_python:IsolatedPythonPlugin', 'submodules = infi.projector.plugins.builtins.submodules:SubmodulePlugin']),
    )

def setup():
    from setuptools import setup as _setup
    from setuptools import find_packages
    SETUP_INFO['packages'] = find_packages('src')
    _setup(**SETUP_INFO)

if __name__ == '__main__':
    setup()




from contextlib import contextmanager
from infi.projector.plugins import CommandPlugin
from infi.projector.helper import assertions
from infi.projector.helper.utils import open_buildout_configfile, commit_changes_to_buildout
from infi.projector.helper.utils.package_sets import PackageDataSet
from textwrap import dedent
from logging import getLogger

logger = getLogger(__name__)

USAGE = """
Usage:
    projector submodule list
    projector submodule add <name> <repository> <rev> [--commit-changes] [--use-setup-py]
    projector submodule remove <name> [--commit-changes]

Options:
    <name>                  name of submodule to add/remove
    <rev>                   remote branch name (must start with origin) or commit hash, e.g. origin/master
    --use-setup-py          add the setup.py of the submodule to the buildout environment
"""

class SubmodulePlugin(CommandPlugin):
    def get_docopt_string(self):
        return USAGE

    def get_command_name(self):
        return 'submodule'

    @assertions.requires_repository
    def parse_commandline_arguments(self, arguments):
        methods = [self.list, self.add, self.remove]
        [method] = [method for method in methods
                    if arguments.get(method.__name__.replace('_', '-'))]
        self.arguments = arguments
        method()

    def get_submodule_sections(self):
        with open_buildout_configfile() as buildout:
            sections = [section for section in buildout.sections()
                        if buildout.has_option(section, "recipe") and \
                        buildout.get(section, "recipe") == "zerokspot.recipe.git"]
            return sections

    def list(self):
        from pprint import pprint
        pprint(self.get_submodule_sections())

    def add(self):
        with open_buildout_configfile(write_on_exit=True) as buildout:
            name = self.arguments.get("<name>")
            if name  not in self.get_submodule_sections():
                buildout.add_section(name)
            repository = self.arguments.get("<repository>")
            rev = self.arguments.get("<rev>")
            buildout.set(name, "recipe", "zerokspot.recipe.git")
            buildout.set(name, "repository", repository)
            buildout.set(name, "rev", rev)
            buildout.set(name, "newest", "true")
            if self.arguments.get("use-setup-py"):
                where_to_look_for_setup_py = set(buildout.get("buildout", "develop").split())
                where_to_look_for_setup_py.add(name)
                buildout.set("buildout", "develop", ' '.join(where_to_look_for_setup_py))
        if self.arguments.get("--commit-changes", False):
            commit_message = "Adding git submodule {}".format(name)
            commit_changes_to_buildout(commit_message)

    def remove(self):
        with open_buildout_configfile(write_on_exit=True) as buildout:
            name = self.arguments.get("<name>")
            if name in self.get_submodule_sections():
                buildout.remove_section(name)
                where_to_look_for_setup_py = set(buildout.get("buildout", "develop").split())
                where_to_look_for_setup_py.discard(name)
                buildout.set("buildout", "develop", ' '.join(where_to_look_for_setup_py))
        if self.arguments.get("--commit-changes", False):
            commit_message = "Removing git submodule {}".format(name)
            commit_changes_to_buildout(commit_message)
